# -*- coding: utf-8 -*-

"""player - Main moveable character in the game."""

##Python imports##
from uuid import uuid4
##Panda3D imports##
from direct.showbase.DirectObject import DirectObject
from direct.fsm.FSM import FSM
from panda3d.core import NodePath, VBase3, CollisionRay
##Local imports##
from .engine import Engine


class Player(object, FSM, DirectObject):

    keys = ('w', 'w-up', 's', 's-up', 'a', 'a-up', 'd', 'd-up')

    def __init__(self, name='Malka', height=50):
        FSM.__init__(self, 'PlayerFSM')
        self.name = name
        self.height = height
        self.node = Engine.base.render.attachNewNode('PlayerNode')
        self.head = self.node.attachNewNode('PlayerHeadNode')
        self.head.setZ(self.height)
        self.keyFlags = {'forward': 0, 'backward': 0, 'strafeLeft': 0,
                          'strafeRight': 0}
        self.walkSpeed = 175
        self.haveMouse = False
        self.focusedObject = None
        self.setupCamera()
        self.setupCollisions()
        self.request('Movement')

    def setupCamera(self):
        Engine.base.camera.reparentTo(self.head)
        Engine.base.camera.clearTransform()
        Engine.base.camLens.setFov(90)

    def use(self, object):
        if not self.focusedObject and hasattr(object, 'setActiveStatue'):
            self.focusedObject = object
            Engine.base.messenger.send('ControlPanel')
            self.request('Controller')

    def setupCollisions(self):
        # Give it a sphere and a ray
        self.sphere = self.node.attachCollisionSphere(uuid4().hex, 0, 0, 15, 12,
                                                      0b1, 0b0)
        self.ray = self.node.attachCollisionRay(uuid4().hex, 0, 0, 0, 0, 0, -1,
                                                0b1, 0b0)
        #self.sphere.show()
        #self.ray.show()
        Engine.pusher.addCollider(self.sphere, self.node)
        Engine.base.cTrav.addCollider(self.sphere, Engine.pusher)
        Engine.gravity.addCollider(self.ray, self.node)
        Engine.base.cTrav.addCollider(self.ray, Engine.gravity)

    def _setKey(self, key, value):
        assert key in self.keyFlags
        self.keyFlags[key] = value
        if (key == 'forward' or key == 'backward' or key == 'strafeLeft' or \
            key == 'strafeRight'):
            if value:
                Engine.base.messenger.send('walking')
            else:
                if not self.keyFlags['forward'] and \
                   not self.keyFlags['backward'] and \
                   not self.keyFlags['strafeLeft'] and \
                   not self.keyFlags['strafeRight']:
                    Engine.base.messenger.send('nowalking')

    def _mouseLook(self, task):
        cursorPos = Engine.base.win.getPointer(0)
        cursorPosX = cursorPos.getX()
        cursorPosY = cursorPos.getY()
        if not Engine.centreCursor():
            self.haveMouse = False
            return task.cont
        elif not self.haveMouse:
            self.haveMouse = True
            return task.cont
        headPitch = self.head.getP()
        self.node.setH(self.node.getH() - (cursorPosX - \
                       Engine.windowCentreX) * 0.1)
        self.head.setP(min(max(headPitch - (cursorPosY - \
                       Engine.windowCentreY) * 0.1, -66), 75))
        return task.cont

    def _move(self, task):
        dt = Engine.globalClock.getDt()
        if self.keyFlags['forward']:
            self.node.setFluidPos(self.node,
                                  VBase3(0, 1, 0) * self.walkSpeed * dt)
        if self.keyFlags['backward']:
            self.node.setFluidPos(self.node,
                                  VBase3(0, -1, 0) * self.walkSpeed * dt)
        if self.keyFlags['strafeLeft']:
            self.node.setFluidPos(self.node,
                                  VBase3(-1, 0, 0) * self.walkSpeed * dt)
        if self.keyFlags['strafeRight']:
            self.node.setFluidPos(self.node,
                                  VBase3(1, 0, 0) * self.walkSpeed * dt)
        return task.cont

    def enterMovement(self):
        Engine.toggleCursor(False)
        Engine.centreCursor()
        self.setupCamera()
        if self.isAccepting('escape'):
            self.ignore('escape')
        if self.focusedObject and hasattr(self.focusedObject,
                                          'setActiveStatue'):
            controller = self.focusedObject
            controller.ignoreAll()
            activeStatue = controller.platform.statues[controller.activeStatue]
            activeStatue.node.clearColorScale()
            self.focusedObject = None
        self.accept('w', self._setKey, ['forward', 1])
        self.accept('w-up', self._setKey, ['forward', 0])
        self.accept('s', self._setKey, ['backward', 1])
        self.accept('s-up', self._setKey, ['backward', 0])
        self.accept('a', self._setKey, ['strafeLeft', 1])
        self.accept('a-up', self._setKey, ['strafeLeft', 0])
        self.accept('d', self._setKey, ['strafeRight', 1])
        self.accept('d-up', self._setKey, ['strafeRight', 0])
        #TODO: Replace this key-based state switch with external triggers.
        Engine.base.accept('escape', Engine.exit)

        Engine.base.taskMgr.doMethodLater(0.5, self._mouseLook, 'MouseLookTask')
        Engine.base.taskMgr.doMethodLater(0.5, self._move, 'PlayerMoveTask')
        self.haveMouse = False

    def enterController(self):
        Engine.base.taskMgr.remove('MouseLookTask')
        Engine.base.taskMgr.remove('PlayerMoveTask')
        for key in Player.keys:
            self.ignore(key)
        for key in self.keyFlags:
            self.keyFlags[key] = 0
        if self.isAccepting('space'):
            print("Ignoring <space>")
            self.ignore('space')
        self.haveMouse = False
        Engine.base.ignore('escape')
        self.acceptOnce('escape', self.request, ['Movement'])
        self.acceptOnce('DoorOpen', self.request, ['Movement'])

