# -*- coding: utf-8 -*-

__all__ = ["Reflector", "BeamOrigin"]

"""lightbeam - Handles the ray of sunlight."""

##Panda3D imports##
from panda3d.core import Point3, Vec3, GeomVertexReader, CollisionHandlerQueue
from panda3d.core import TextureStage, TransparencyAttrib, CardMaker
##Local imports##
from src.engine import Engine


class Reflector(object):

    # This must be fixed in the beam model too
    divergence = 2.0

    def __init__(self, nodePath):
        self.task = Engine.base.taskMgr.add(self._task, 'collisions')
        self.other = None
        self.node = nodePath
        self.node.flattenStrong()
        self.node.setDepthOffset(1)
        self.node.setCollideMask(0b10)
        # WARNING circular dependency
        self.node.setPythonTag('reflector', self)

        geom = self.node.node().getGeom(0)
        vdata = geom.getVertexData()
        vertex = GeomVertexReader(vdata, 'vertex')
        self.vertices = []
        while not vertex.isAtEnd():
            self.vertices.append(vertex.getData3f())
        assert len(self.vertices) == 4
        self.centre = Point3(self.vertices[0] + self.vertices[1] + \
                             self.vertices[2] + self.vertices[3]) * 0.25
        vertDiff = self.vertices[1] - self.vertices[0]
        self.normal = Vec3(vertDiff.cross(self.vertices[2] - self.vertices[0]))
        self.normal.normalize()
        self.length = (self.vertices[1] - self.vertices[0]).length()
        self.ray = self.node.attachCollisionRay('ray', self.centre[0],
                                                self.centre[1], self.centre[2],
                                                self.normal[0], self.normal[1],
                                                self.normal[2], 0b11, 0b00)
        self.handler = CollisionHandlerQueue()
        Engine.base.cTrav.addCollider(self.ray, self.handler)

        self.lightBeam = Engine.base.loader.loadModel('lightbeam')
        self.lightBeam.reparentTo(self.node)
        self.lightBeam.lookAt(Point3(self.normal))
        self.lightBeam.setPos(self.centre)
        self.lightBeam.setScale(self.length, 1, self.length)
        ts = TextureStage('lightbeam')
        ts.setMode(TextureStage.MModulateGlow)
        self.brightSpotTex = Engine.base.loader.loadTexture('brightspot.png')
        self.lightBeam.setTexture(self.brightSpotTex)
        self.lightBeam.setTransparency(TransparencyAttrib.MDual, 2)
        self.lightBeam.setLightOff(1)
        self.lightBeam.setAlphaScale(0.5)
        self.lightBeam.hide()

        cm = CardMaker('brightspot')
        cm.setFrame(-0.5, 0.5, -0.5, 0.5)
        self.brightSpot = Engine.base.render.attachNewNode(cm.generate())
        self.brightSpot.setP(180)
        self.brightSpot.flattenStrong()
        self.brightSpot.setDepthOffset(2)
        self.brightSpot.setAlphaScale(0.9)
        self.brightSpot.setLightOff(1)
        self.brightSpot.setTransparency(TransparencyAttrib.MAlpha, 2)
        self.brightSpot.setScale(self.length * self.divergence)
        ts = TextureStage('brightspot')
        ts.setMode(TextureStage.MModulateGlow)
        self.brightSpot.setTexture(self.brightSpotTex)
        self.brightSpot.hide()

        self.victoryTargets = {}
        self.victoryTargetPriorities = {}
        self.victory = False

    def reflect(self, vector, length):
        diff = self.normal + vector
        newVec = self.normal + diff
        self.ray.node().modifySolid(0).setDirection(newVec)
        self.lightBeam.lookAt(self.lightBeam.getPos() + newVec)
        self.lightBeam.show()
        self.lightBeam.setSx(length)
        self.lightBeam.setSz(length)
        self.brightSpot.setScale(length * self.divergence)

    def disable(self):
        self.lightBeam.hide()
        self.brightSpot.hide()
        if self.other:
            self.other.lightBeam.hide()
            self.other.brightSpot.hide()

    def setVictoryTarget(self, target):
        self.victoryTargets[target] = False
        target.setCollideMask(0b10)
        priority = target.getNetPythonTag('priority')
        self.victoryTargetPriorities[priority] = target
        self.victoryTargetPriorities[target] = priority

    def _task(self, task):
        self.handler.sortEntries()
        entry = None
        for e in self.handler.getEntries():
            if e.getIntoNodePath() != self.node:
                entry = e
                break
        if entry and not self.lightBeam.isHidden():
            entryNode = entry.getIntoNodePath()
            if entryNode in self.victoryTargets and not self.victory:
                if not self.victoryTargets[entryNode]:
                    priority = self.victoryTargetPriorities[entryNode]
                    if priority is 0:
                        self.victoryTargets[entryNode] = True
                        entryNode.hide()
                    elif priority is 1:
                        if self.victoryTargets[self.victoryTargetPriorities[0]]:
                            self.victoryTargets[entryNode] = True
                            entryNode.hide()
                    elif priority is 2:
                        if self.victoryTargets[self.victoryTargetPriorities[0]] and \
                           self.victoryTargets[self.victoryTargetPriorities[1]]:
                            self.victoryTargets[entryNode] = True
                            entryNode.hide()
                        else:
                            for target in self.victoryTargets:
                                self.victoryTargets[target] = False
                                target.show()
                    if not False in list(self.victoryTargets.values()):
                        Engine.base.messenger.send('BalancingPuzzleVictory')
                        self.victory = True
            self.brightSpot.lookAt(self.brightSpot.getPos() + \
                                   entry.getSurfaceNormal(Engine.base.render))
            self.brightSpot.setPos(entry.getSurfacePoint(Engine.base.render))
            self.lightBeam.setSy((entry.getSurfacePoint(self.node) - \
                                  self.lightBeam.getPos()).length())
            self.brightSpot.show()
            newOther = entry.getIntoNodePath().getPythonTag('reflector')
            if self.other != newOther:
                if self.other:
                    self.other.disable()
                self.other = newOther
                if self.other:
                    otherRelVec = self.other.node.getRelativeVector(self.node,
                                                                    self.normal)
                    self.other.reflect(otherRelVec,
                                       self.length * self.divergence)
        else:
            self.brightSpot.hide()
            if self.other:
                 self.other.disable()
                 self.other = None
        return task.cont


class BeamOrigin(Reflector):
    """Inherits from Reflector because they share functionality"""

    def __init__(self, nodePath):
        Reflector.__init__(self, nodePath)
        self.node.setLightOff(1)
        self.node.setCollideMask(0b00)
        self.lightBeam.show()
        self.reflectors = []

    def addReflector(self, reflector):
        if not isinstance(reflector, Reflector):
            return self.addReflector(Reflector(reflector))
        self.reflectors.append(reflector)

