# -*- coding: utf-8 -*-

"""world - Sets up and manages everything in the game world."""

##Python imports##
from sys import exit
from random import random, choice
##Panda3D imports##
from direct.filter.CommonFilters import CommonFilters
from direct.showbase.DirectObject import DirectObject
from direct.showbase.Transitions import Transitions
from direct.interval.LerpInterval import LerpPosInterval
from direct.interval.FunctionInterval import Func
from direct.interval.IntervalGlobal import Parallel, Sequence, Wait
from direct.gui.OnscreenText import OnscreenText
from panda3d.core import AmbientLight, PointLight
from panda3d.core import VBase3, VBase4, Point3
from panda3d.core import TransparencyAttrib
from panda3d.core import CardMaker
##Local imports##
from .engine import Engine
from .player import Player
from .audio import Audio
from src.puzzles.balancing.platform import Platform
from src.puzzles.balancing.statue import Statue
from src.puzzles.balancing.lightbeam import *
from src.puzzles.balancing.controller import Controller


class World(object, DirectObject):

    def __init__(self):
        self.instructionText = None
        self.victorious = False

        self.setupMap()
        self.setupPlayer()
        self.setupPlatform()
        self.setupBeam()
        self.setupControllers()
        self.setupLighting()

    def setupMap(self):
        self.map = Engine.base.loader.loadModel('room')
        self.map.reparentTo(Engine.base.render)
        # Because somehow the map has alpha enabled, and that interferes with
        # the light beam stuff.
        self.map.setTransparency(TransparencyAttrib.MOff, 1)

        self.entranceStone = self.map.find('**/entrance_stone').getParent()
        self.stoneUpPos = self.entranceStone.getPos()
        self.stoneDownPos = Point3(self.entranceStone.getPos()) # Duplicate
        self.stoneDownPos.setZ(0)
        self.stoneDownSoundInterval = Func(Audio.playSound, 'door_close', 0.8)
        self.stoneUpSoundInterval = Func(Audio.playSound, 'door_open', 0.8)
        self.entranceStoneDropIval = LerpPosInterval(self.entranceStone, 1.2,
                                                     self.stoneDownPos)
        self.entranceStoneRaiseIval = LerpPosInterval(self.entranceStone, 1.2,
                                                      self.stoneUpPos)
        self.entranceStoneDrop = Parallel(self.entranceStoneDropIval,
                                          self.stoneDownSoundInterval)
        self.entranceStoneRaise = Parallel(self.entranceStoneRaiseIval,
                                           self.stoneUpSoundInterval)

    def checkPlayerAtGate(self, task):
        if not self.victorious:
            if self.player.node.getZ() < -1 and \
               self.player.node.getDistance(self.entranceStone) > 275:
                self.entranceStoneDrop.start()
                self.acceptOnce('BalancingPuzzleVictory', self.victory)
                # Turn the player around to look at the closing door.
                playerTransform = self.player.node.getTransform()
                self.player.node.lookAt(self.entranceStone.getParent(),
                                        self.stoneDownPos)
                newTransform = self.player.node.getTransform()
                self.player.node.setTransform(playerTransform)
                playerTurn = self.player.node.hprInterval(0.5,
                                                          newTransform.getHpr(),
                                                          blendType='easeInOut')
                playerFreeze = Func(self.player.request, 'Controller')
                playerRestore = Func(self.player.request, 'Movement')
                musicStart = Func(self.playMusic)
                Sequence(Wait(0.2), playerFreeze, playerTurn, Wait(0.4),
                         musicStart, playerRestore).start()
                return task.done
        else:
            if self.player.node.getZ() > 0:
                self.player.request('Cursor')
                cm = CardMaker('black')
                cm.setColor(0, 0, 0, 1)
                cm.setFrame(-1, 1, -1, 1)
                card = Engine.base.render2d.attachNewNode(cm.generate())
                card.setTransparency(TransparencyAttrib.MAlpha, 3)
                card.setColorScale(1, 1, 1, 0.0)
                card.colorScaleInterval(2.0, VBase4(1, 1, 1, 1),
                                        startColorScale=VBase4(1, 1, 1, 0)).start()
                font = Engine.fonts['Ink Calligraphy']
                congratsText = OnscreenText('Congratulations!\nPress ' + \
                                            'Escape to exit',
                                            parent=Engine.base.aspect2dp,
                                            pos=(0, -0.85), scale=0.1,
                                            fg=(1, 1, 1, 1), font=font)
                return task.done
        return task.cont

    def victory(self):
        Sequence(self.entranceStoneRaise,
                 Func(Engine.base.messenger.send, 'DoorOpen')).start()
        self.victorious = True
        self.addTask(self.checkPlayerAtGate, 'PlayerAtGateTask')

    def checkPlayerNearControl(self, task):
        if self.player.getCurrentOrNextState() == 'Movement':
            if self.player.node.getDistance(self.controller.node) < 75:
                if not self.instructionText:
                    font = Engine.fonts['Ink Calligraphy']
                    self.instructionText = OnscreenText('Placeholder',
                                                        pos=(0, -0.75),
                                                        scale=0.1,
                                                        fg=(1, 1, 1, 1),
                                                        font=font, mayChange=True)
                elif self.instructionText.isHidden():
                    self.instructionText.show()
                self.instructionText['text'] = 'Press Spacebar to use controls.'
                if not self.player.isAccepting('space'):
                    self.player.accept('space', self.player.use, [self.controller])
                    print("Accepting <space>")
            else:
                if self.instructionText and not self.instructionText.isHidden():
                    self.instructionText.hide()
                if self.player.isAccepting('space'):
                    self.player.ignore('space')
                    print("Ignoring <space>")
        return task.cont

    def setupPlayer(self):
        self.player = Player('Malka')
        playerStart = self.map.find('**/player_start').getTransform()
        self.player.node.setTransform(playerStart)
        self.player.node.setScale(1)
        self.accept('ControlPanel', self.fixCameraView)
        Audio.addSoundEvent('walking', 'footsteps', loop=True, volume=0.6)
        Audio.addSoundEvent('nowalking', 'footsteps', playRate=0.0)
        Audio.playSound('breathing', loop=True, volume=0.5)

        self.addTask(self.checkPlayerAtGate, 'PlayerAtGateCheckTask')
        self.addTask(self.checkPlayerNearControl, 'PlayerNearControlCheckTask')

    def fixCameraView(self):
        self.instructionText['text'] = 'Up/Down Arrows to select a statue. ' + \
                                       'Left/Right Arrows to rotate it.\n' + \
                                       'Comma(,)/Period(.) to rotate the platform.\n' + \
                                       'Escape(Esc) to return.'
        fixedCam = self.map.find('**/camera_position')
        Engine.base.camera.reparentTo(fixedCam)
        Engine.base.camera.clearTransform()
        Engine.base.camLens.setFov(85)
        self.controller.setupControls()

    def setupPlatform(self):
        platform = self.map.find('**/platform_bottom')
        self.platform = Platform(platform, self.player)
        for i in range(4):
            newStatue = Statue(i, self.map.find('**/statue' + str(i + 1)),
                               (random() * 4)**(random() * 3))
            self.platform.statues.append(newStatue)
        self.platform.shufflePositions()

    def setupBeam(self):
        sequenceQueue = [0, 1, 2]
        target1 = self.map.find('**/beam_target')
        priority = choice(sequenceQueue)
        sequenceQueue.remove(priority)
        target1.setPythonTag('priority', priority)
        priority = choice(sequenceQueue)
        sequenceQueue.remove(priority)
        target2 = self.map.find('**/beam_target2')
        target2.setScale(2)
        target2.setPythonTag('priority', priority)
        priority = choice(sequenceQueue)
        sequenceQueue.remove(priority)
        target3 = self.map.find('**/beam_target3')
        target3.setPythonTag('priority', priority)
        target1.setTransparency(TransparencyAttrib.MBinary, 3)
        target2.setTransparency(TransparencyAttrib.MBinary, 3)
        target3.setTransparency(TransparencyAttrib.MBinary, 3)
        reflector = Reflector(self.map.find('**/beam_reflector'))
        reflector.setVictoryTarget(target1)
        reflector.setVictoryTarget(target2)
        reflector.setVictoryTarget(target3)
        self.lightBeam = BeamOrigin(self.map.find('**/beam_origin'))
        self.lightBeam.addReflector(reflector)

    def setupControllers(self):
        self.controller = Controller(self.map.find('**/control_panel'),
                                     self.platform)

    def setupLighting(self):
        #TODO: Move the lighting to another module.
        self.alight = AmbientLight('AmbientLight')
        self.alight.setColor(VBase4(0.6, 0.6, 0.55, 1))
        self.alightNode = Engine.base.render.attachNewNode(self.alight)
        Engine.base.render.setLight(self.alightNode)
        self.plight = PointLight('PointLight')
        self.plight.setColor(VBase4(1, 1, 0.95, 1))
        self.plight.setSpecularColor(VBase4(0.4, 0.4, 0.35, 1))
        #self.plight.setAttenuation(VBase3(0.1, 0.00001, 0.000001))
        self.plightNode = self.lightBeam.node.attachNewNode(self.plight)
        Engine.base.render.setLight(self.plightNode)
        self.cf = CommonFilters(Engine.base.win, Engine.base.cam)
        self.cf.setBloom(desat=0.0, intensity=2.0, size='medium')
        Engine.base.render.setShaderAuto()

    def playMusic(self):
        Audio.blacklistTrack('Menu Music')
        Audio.playTrack('Amoris Immortalis', volume=0.5)
        Audio.eventHandler.accept('finishedAmoris Immortalis',
                                  Audio.playRandomTrack)

